/*
 * Copyright © 2021-2023 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of ipspiel24-VierConnects-core.
 *
 * ipspiel24-VierConnects-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * ipspiel24-VierConnects-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * ipspiel24-VierConnects-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.ipspiel24.VierConnects.core.domain.impl;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.regex.Pattern;

import de.fhdw.gaming.core.domain.GameBuilder;
import de.fhdw.gaming.core.domain.GameBuilderFactory;
import de.fhdw.gaming.core.domain.GameException;
import de.fhdw.gaming.core.domain.Strategy;
import de.fhdw.gaming.core.ui.InputProvider;
import de.fhdw.gaming.core.ui.InputProviderException;
import de.fhdw.gaming.core.ui.type.validator.MaxValueValidator;
import de.fhdw.gaming.core.ui.type.validator.MinValueValidator;
import de.fhdw.gaming.core.ui.type.validator.PatternValidator;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsGameBuilder;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsGameBuilderFactory;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsPlayer;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsPlayerBuilder;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsStrategy;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.factory.VierConnectsDefaultStrategyFactoryProvider;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.factory.VierConnectsStrategyFactory;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.factory.VierConnectsStrategyFactoryProvider;
import de.fhdw.gaming.ipspiel24.VierConnects.core.moves.factory.VierConnectsMoveFactory;
import de.fhdw.gaming.ipspiel24.VierConnects.core.moves.impl.VierConnectsDefaultMoveFactory;

/**
 * Implements {@link VierConnectsGameBuilderFactory} by creating a Vier Connects game builder.
 */
public final class VierConnectsGameBuilderFactoryImpl implements VierConnectsGameBuilderFactory {

    /**
     * The number of players.
     */
    private static final int NUMBER_OF_PLAYERS = 2;
    /**
     * Minimum number of rows of the board.
     */
    private static final int MIN_NR_OF_ROWS = 4;
    /**
     * Maximum number of rows of the board.
     */
    private static final int MAX_NR_OF_ROWS = 16;
    /**
     * Minimum number of columns of the board.
     */
    private static final int MIN_NR_OF_COLUMNS = 5;
    /**
     * Maximum number of columns of the board.
     */
    private static final int MAX_NR_OF_COLUMNS = 16;
    /**
     * Smallest allowed maximum computation time per move in seconds.
     */
    private static final int MIN_MAX_COMPUTATION_TIME_PER_MOVE = 1;
    /**
     * Largest allowed maximum computation time per move in seconds.
     */
    private static final int MAX_MAX_COMPUTATION_TIME_PER_MOVE = 3600;

    /**
     * All available Vier Connects strategies.
     */
    private final Set<VierConnectsStrategy> strategies;

    /**
     * Creates an Vier Connects game factory. Vier Connects strategies are loaded by using the
     * {@link java.util.ServiceLoader}.
     * <p>
     * This constructor is meant to be used by the {@link java.util.ServiceLoader}.
     */
    public VierConnectsGameBuilderFactoryImpl() {
        this(new VierConnectsDefaultStrategyFactoryProvider());
    }

    /**
     * Creates an Vier Connects game factory.
     *
     * @param strategyFactoryProvider The {@link VierConnectsStrategyFactoryProvider} for loading Vier Connects
     *                                strategies.
     */
    VierConnectsGameBuilderFactoryImpl(final VierConnectsStrategyFactoryProvider strategyFactoryProvider) {
        final VierConnectsMoveFactory moveFactory = new VierConnectsDefaultMoveFactory();

        final List<VierConnectsStrategyFactory> factories = strategyFactoryProvider.getStrategyFactories();
        this.strategies = new LinkedHashSet<>();
        for (final VierConnectsStrategyFactory factory : factories) {
            this.strategies.add(factory.create(moveFactory));
        }
    }

    @Override
    public String getName() {
        return "VierConnects";
    }

    @Override
    public int getMinimumNumberOfPlayers() {
        return VierConnectsGameBuilderFactoryImpl.NUMBER_OF_PLAYERS;
    }

    @Override
    public int getMaximumNumberOfPlayers() {
        return VierConnectsGameBuilderFactoryImpl.NUMBER_OF_PLAYERS;
    }

    @Override
    public List<? extends Strategy<?, ?, ?>> getStrategies() {
        return new ArrayList<>(this.strategies);
    }

    @Override
    public VierConnectsGameBuilder createGameBuilder(final InputProvider inputProvider) throws GameException {
        try {
            final VierConnectsGameBuilder gameBuilder = new VierConnectsGameBuilderImpl();

            @SuppressWarnings("unchecked")
            final Map<String, Object> gameData = inputProvider
                    .needInteger(
                            VierConnectsGameBuilderFactory.PARAM_NR_OF_ROWS,
                            "Number of rows",
                            Optional.of(VierConnectsGameBuilder.DEFAULT_NR_OF_ROWS),
                            new MinValueValidator<>(VierConnectsGameBuilderFactoryImpl.MIN_NR_OF_ROWS),
                            new MaxValueValidator<>(VierConnectsGameBuilderFactoryImpl.MAX_NR_OF_ROWS))
                    .needInteger(
                            VierConnectsGameBuilderFactory.PARAM_NR_OF_COLUMNS,
                            "Number of columns",
                            Optional.of(VierConnectsGameBuilder.DEFAULT_NR_OF_COLUMNS),
                            new MinValueValidator<>(VierConnectsGameBuilderFactoryImpl.MIN_NR_OF_COLUMNS),
                            new MaxValueValidator<>(VierConnectsGameBuilderFactoryImpl.MAX_NR_OF_COLUMNS))
                    .needInteger(
                            GameBuilderFactory.PARAM_MAX_COMPUTATION_TIME_PER_MOVE,
                            "Maximum computation time per move in seconds",
                            Optional.of(GameBuilder.DEFAULT_MAX_COMPUTATION_TIME_PER_MOVE),
                            new MinValueValidator<>(
                                    VierConnectsGameBuilderFactoryImpl.MIN_MAX_COMPUTATION_TIME_PER_MOVE),
                            new MaxValueValidator<>(
                                    VierConnectsGameBuilderFactoryImpl.MAX_MAX_COMPUTATION_TIME_PER_MOVE))
                    .requestData("Board properties");

            gameBuilder.changeBoardSize((Integer) gameData.get(VierConnectsGameBuilderFactory.PARAM_NR_OF_ROWS),
                    (Integer) gameData.get(VierConnectsGameBuilderFactory.PARAM_NR_OF_COLUMNS));
            gameBuilder.changeMaximumComputationTimePerMove(
                    (Integer) gameData.get(GameBuilderFactory.PARAM_MAX_COMPUTATION_TIME_PER_MOVE));

            final InputProvider firstPlayerInputProvider = inputProvider.getNext(gameData);
            final Map<String, Object> firstPlayerData = this
                    .requestPlayerData(firstPlayerInputProvider, "Player 1", Optional.empty());
            final VierConnectsPlayer firstPlayer = this.createPlayer(gameBuilder.createPlayerBuilder(),
                    firstPlayerData);
            final VierConnectsStrategy firstPlayerStrategy = this.getStrategy(firstPlayerData);
            gameBuilder.addPlayer(firstPlayer, firstPlayerStrategy);

            final InputProvider secondPlayerInputProvider = firstPlayerInputProvider.getNext(firstPlayerData);
            final Map<String, Object> secondPlayerData = this.requestPlayerData(
                    secondPlayerInputProvider,
                    "Player 2",
                    Optional.of(
                            !(Boolean) firstPlayerData.get(VierConnectsGameBuilderFactory.PARAM_PLAYER_USING_CROSSES)));
            final VierConnectsPlayer secondPlayer = this.createPlayer(gameBuilder.createPlayerBuilder(),
                    secondPlayerData);
            final VierConnectsStrategy secondPlayerStrategy = this.getStrategy(secondPlayerData);
            gameBuilder.addPlayer(secondPlayer, secondPlayerStrategy);

            return gameBuilder;
        } catch (final InputProviderException e) {
            throw new GameException(String.format("Creating Vier Connects game was aborted: %s", e.getMessage()), e);
        }
    }

    /**
     * Initialises an Vier Connects player builder.
     *
     * @param inputProvider The input provider.
     * @param title         The title for the UI.
     * @param usingCrosses  If set, determines whether the player uses crosses or noughts. Otherwise, the setting is
     *                      initially set to crosses, and the user can change it at will.
     * @return {@code playerBuilder}.
     * @throws InputProviderException if the operation has been aborted prematurely (e.g. if the user cancelled a
     *                                dialog).
     */
    @SuppressWarnings("unchecked")
    private Map<String, Object> requestPlayerData(final InputProvider inputProvider, final String title,
            final Optional<Boolean> usingCrosses) throws GameException, InputProviderException {

        inputProvider
                .needString(
                        GameBuilderFactory.PARAM_PLAYER_NAME,
                        "Name",
                        Optional.empty(),
                        new PatternValidator(Pattern.compile("\\S+(\\s+\\S+)*")))
                .needBoolean(
                        VierConnectsGameBuilderFactory.PARAM_PLAYER_USING_CROSSES,
                        "Uses Red",
                        Optional.of(Boolean.TRUE))
                .needObject(GameBuilderFactory.PARAM_PLAYER_STRATEGY, "Strategy", Optional.empty(), this.strategies);

        if (usingCrosses.isPresent()) {
            inputProvider.fixedBoolean(VierConnectsGameBuilderFactory.PARAM_PLAYER_USING_CROSSES, usingCrosses.get());
        }

        return inputProvider.requestData(title);
    }

    /**
     * Creates an Vier Connects player builder.
     *
     * @param playerBuilder The player builder.
     * @param playerData    The requested player data.
     * @return {@link VierConnectsPlayer}.
     * @throws InputProviderException if the operation has been aborted prematurely (e.g. if the user cancelled a
     *                                dialog).
     */
    private VierConnectsPlayer createPlayer(final VierConnectsPlayerBuilder playerBuilder,
            final Map<String, Object> playerData) throws GameException, InputProviderException {

        return playerBuilder.changeName((String) playerData.get(GameBuilderFactory.PARAM_PLAYER_NAME))
                .changeUsingCrosses((Boolean) playerData.get(VierConnectsGameBuilderFactory.PARAM_PLAYER_USING_CROSSES))
                .build();
    }

    /**
     * Returns an Vier Connects strategy.
     *
     * @param playerData The requested player data.
     * @return The Vier Connects strategy.
     */
    private VierConnectsStrategy getStrategy(final Map<String, Object> playerData) {
        return (VierConnectsStrategy) playerData.get(GameBuilderFactory.PARAM_PLAYER_STRATEGY);
    }
}
