/*
 * Copyright © 2021-2023 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of ipspiel24-VierConnects-core.
 *
 * ipspiel24-VierConnects-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * ipspiel24-VierConnects-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * ipspiel24-VierConnects-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.ipspiel24.VierConnects.core.domain.impl;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

import de.fhdw.gaming.core.domain.DefaultGame;
import de.fhdw.gaming.core.domain.DefaultObserverFactoryProvider;
import de.fhdw.gaming.core.domain.Game;
import de.fhdw.gaming.core.domain.GameBuilder;
import de.fhdw.gaming.core.domain.GameException;
import de.fhdw.gaming.core.domain.ObserverFactoryProvider;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsGameBuilder;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsPlayer;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsPlayerBuilder;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsState;
import de.fhdw.gaming.ipspiel24.VierConnects.core.domain.VierConnectsStrategy;
import de.fhdw.gaming.ipspiel24.VierConnects.core.moves.VierConnectsMove;
import de.fhdw.gaming.ipspiel24.VierConnects.core.moves.impl.AbstractVierConnectsMove;

/**
 * Implements {@link VierConnectsGameBuilder}.
 */
final class VierConnectsGameBuilderImpl implements VierConnectsGameBuilder {

    /**
     * The {@link ObserverFactoryProvider}.
     */
    private ObserverFactoryProvider observerFactoryProvider;

    /**
     * The player using crosses.
     */
    private Optional<VierConnectsPlayer> crossesPlayer;

    /**
     * The strategy of the player using crosses.
     */
    private Optional<VierConnectsStrategy> crossesPlayerStrategy;

    /**
     * The player using noughts.
     */
    private Optional<VierConnectsPlayer> noughtsPlayer;

    /**
     * The strategy of the player using noughts.
     */
    private Optional<VierConnectsStrategy> noughtsPlayerStrategy;

    /**
     * The maximum computation time per move in seconds.
     */
    private int maxComputationTimePerMove;

    /**
     * The number of rows of the board.
     */
    private int nrOfRows;

    /**
     * The number of columns of the board.
     */
    private int nrOfColumns;

    /**
     * Creates an VierConnects game builder.
     */
    VierConnectsGameBuilderImpl() {
        this.observerFactoryProvider = new DefaultObserverFactoryProvider();
        this.crossesPlayer = Optional.empty();
        this.crossesPlayerStrategy = Optional.empty();
        this.noughtsPlayer = Optional.empty();
        this.noughtsPlayerStrategy = Optional.empty();
        this.maxComputationTimePerMove = GameBuilder.DEFAULT_MAX_COMPUTATION_TIME_PER_MOVE;
        this.nrOfRows = VierConnectsGameBuilder.DEFAULT_NR_OF_ROWS;
        this.nrOfColumns = VierConnectsGameBuilder.DEFAULT_NR_OF_COLUMNS;
    }

    @Override
    public VierConnectsPlayerBuilder createPlayerBuilder() {
        return new VierConnectsPlayerBuilderImpl();
    }

    @Override
    public VierConnectsGameBuilder addPlayer(final VierConnectsPlayer player, final VierConnectsStrategy strategy)
            throws GameException {

        if (player.isUsingCrosses() && this.crossesPlayer.isEmpty()) {
            this.crossesPlayer = Optional.of(Objects.requireNonNull(player, "player"));
            this.crossesPlayerStrategy = Optional.of(Objects.requireNonNull(strategy, "crossesPlayerStrategy"));
        } else if (!player.isUsingCrosses() && this.noughtsPlayer.isEmpty()) {
            this.noughtsPlayer = Optional.of(Objects.requireNonNull(player, "player"));
            this.noughtsPlayerStrategy = Optional.of(Objects.requireNonNull(strategy, "noughtsPlayerStrategy"));
        } else {
            throw new GameException(
                    String.format(
                            "Adding player %s is not allowed as a player using the same type of marks "
                                    + "has already been added.",
                            player));
        }
        return this;
    }

    @Override
    public VierConnectsGameBuilder changeMaximumComputationTimePerMove(final int newMaxComputationTimePerMove) {
        this.maxComputationTimePerMove = newMaxComputationTimePerMove;
        return this;
    }

    @Override
    public VierConnectsGameBuilder changeBoardSize(final int newNrOfRows, final int newNrOfColumns) {
        this.nrOfRows = newNrOfRows;
        this.nrOfColumns = newNrOfColumns;
        return this;
    }

    @Override
    public VierConnectsGameBuilder changeObserverFactoryProvider(
            final ObserverFactoryProvider newObserverFactoryProvider) {
        this.observerFactoryProvider = newObserverFactoryProvider;
        return this;
    }

    @Override
    public Game<VierConnectsPlayer, VierConnectsState, VierConnectsMove, VierConnectsStrategy> build(final int id)
            throws GameException, InterruptedException {
        if (!this.crossesPlayer.isPresent() || !this.noughtsPlayer.isPresent()) {
            throw new GameException("An Vier Connects game needs two players.");
        }

        final VierConnectsBoardImpl board = new VierConnectsBoardImpl(this.nrOfRows, this.nrOfColumns);
        final VierConnectsState initialState = new VierConnectsStateImpl(
                board,
                this.crossesPlayer.get(),
                this.noughtsPlayer.get(),
                true);

        final Map<String, VierConnectsStrategy> strategies = new LinkedHashMap<>();
        strategies.put(initialState.getCrossesPlayer().getName(), this.crossesPlayerStrategy.orElseThrow());
        strategies.put(initialState.getNoughtsPlayer().getName(), this.noughtsPlayerStrategy.orElseThrow());
        return new DefaultGame<>(
                id,
                initialState,
                strategies,
                this.maxComputationTimePerMove,
                AbstractVierConnectsMove.class::isInstance,
                new VierConnectsMoveGeneratorImpl(),
                this.observerFactoryProvider);
    }
}
